/******************************************************************************
*
*	TITLE:		Autonomous.c 
*                          
*
*	DATE:		15-Feb-2007
*
*	AUTHOR:	1511
*			
*
*	COMMENTS:	This file contains the code for one of the autonomous modes.
*
*				Specifically, this approach attempts to drive straight into the
*           diamond plate (hopefully with little or no angle)
*
*
*******************************************************************************
*
*	CHANGE LOG:
*
******************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <math.h>
#include "tracking.h"
#include "Motion.h"

/****************  Defines for this file only *********************/
// Yea I know people hate this but at least it is readable and
// clearly defined.
#define TRUE 1
#define FALSE 0

#define LEFT 0
#define RIGHT 1

#define RAMPING   TRUE
#define NO_RAMPING FALSE	

// 12 inches per foot -- think in feet and deal in inches
#define FT2IN  12       
#define PI ((double)(3.14159))

// *************  Adjustable / tunable items
// ********* Note -- Although the values are listed as FT/SEC and
// *********         then converted to IN/SEC, the exact speed
// *********         is not important except for turns, the 
// *********         slow wheel must be half the fast wheel
// *********         for the turning control to work correctly
// FT/SEC The speed for final tracking to the diamond plate 
#define ROBOT_TRK_SPD	 ((short)(4 * FT2IN))  
// Inches - Speed delta for L/R tracking
#define TRK_RATE           2                   
// FT/SEC Speed of the robot till the final approach
#define ROBOT_SPEED		 ((short)(6 * FT2IN))  
// Basically will have 1 side turn at half speed of the other
#define ROBOT_TURN_RATIO ((short)(2 * FT2IN))  
// FT - Distance we want to approach the diamond plate straight on
#define APPROACH_DIST	 ((short)(7 * FT2IN))  
// FT - How far to travel before calculating the approach
//      Will want to get up to speed during the START_DIST
#define START_DIST		 ((short)(2 * FT2IN))  


// ************* Not expected to be adjusted
// FT - Length of rack leg from center of rack
#define RACK_LEG_LEN	    ((short)(4 * FT2IN))	
// Entered as inches for dist from wheel center to wheel center XXXxxxx
#define WHEEL2WHEEL_DIST  24   
#define TURN_CIRCUMFERANCE  ((short)(2 * WHEEL2WHEEL_DIST * PI))
 // 17.6 is the I/S for 1mph (probably 1 will be best value)
#define MAX_SPD_RAMP   1      


// The states this mode goes through to achieve the goal
enum
{
    MOVE_FROM_START,
    TURN_TO_ALIGN,       
    MOVE_TO_ALIGN,
    TURN_TO_APPROACH, 
    DELAY_FOR_CAMERA,       
    MOVE_TO_APPROACH,       
    DROP_TUBE,              
    ENDING_MOVE, 
	 END_OF_STATE_TABLE
} ControlState;



short DistTraveledIn = 0;       // Distance traveled in inches
short LeftSpeedControl = 0;     // Commanded Speed in Inch/Sec of Left wheel
short RightSpeedControl = 0;    // Commanded Speed in Inch/Sec of Right wheel

// Alignment angle and distance are used to guide the robot to the approach vector
// The approach angle is used to turn from the alignment position to the approach
// position and then the approach distance can be used to get to the rack (or the light 
// can be used)
short AlignDistIn = 0; 
short AlignAngle = 0;               // will be in integer degrees
short ApproachDistIn = 0; 
short ApproachAngle = 0;            // will be in integer degrees

unsigned long CurrentTachReadingL;
unsigned long CurrentTachReadingR;

// xxxx should be switch configured
int LeftRightSelect = RIGHT;        // If the camera has to lights this picks which one to go to
int DelayForCamera;                 // Used if we  wish to wait some number of loops for camera to track

// Forward declarations for the compiler
void  AutonomousControl(void);
void  InitDistance();
short CountDistance(short LeftRightWheel);
void  MoveForward(short InchPerSecL, short InchPerSecR, short Ramping);
void  LeftWheelSpeed(short InchPerSec);
void  RightWheelSpeed(short InchPerSec);
void  TrackCameraHorizontal(short LeftRightSelect);
void  TrackCameraVertical();
void  CalcMoves();
short WideTurn(short TurnAngle);
//  These may turn into external declarations
short R_GetCamHorizLightPos();
short ReleaseGripper();
short R_GetCamDist();
float R_GetCamHorizAngle();
void  R_GripRelease();
short R_GripIsReleased();


void     R_SetLeftWheelSpeed(InchPerSec);
void     R_SetRightWheelSpeed(InchPerSec);
unsigned long R_GetTach(short LeftRight);




int main()
{
   InitDistance();
   CountDistance(LEFT);
   AutonomousControl();
   MoveForward(ROBOT_SPEED,ROBOT_SPEED,RAMPING);   // Move straight with rampinp
   return(0);
}


/****************************************************
 InitDistance -- 

   Resets the distance traveled to 0.  Expected
   autonomous use is whenever a you know how far
   you want to travel, you can set this to zero
   and then wait till it gets to the desired 
   distance.
*****************************************************/
void InitDistance()
{
   DistTraveledIn = 0;
   CurrentTachReadingL = R_GetTach(LEFT);
   CurrentTachReadingR = R_GetTach(RIGHT);

}

/****************************************************
 CountDistance -- 

   Converts tach pulses to distance traveled for the 
   side requested -- In inches!
   NOTE -- can not use distance from both sides because
   this function updates both counters so use one side
   for distance on straight aways and the slower one
   for turns.
*****************************************************/
short CountDistanceIn(short LeftRight)
{
   unsigned long NewTachReadingL;
   unsigned long NewTachReadingR;
   unsigned long tmp;

   NewTachReadingL = R_GetTach(LEFT);
   NewTachReadingR = R_GetTach(RIGHT);

   if (LeftRight == LEFT)
   {
      // Check to see if counter wrapped around
      if (NewTachReadingL < CurrentTachReadingL)
      {
         tmp = (0xFFFFFFFFUL - CurrentTachReadingL + 
                            NewTachReadingL) * TICKS_PER_INCH;
         DistTraveledIn += (short)tmp;
         if (tmp > 2000)
            printf("CountDistanceIn -- bad calc of distance %ld\n",tmp);
      }
      else  // Counter not wrapped around so straight calc
      {
         tmp = (NewTachReadingL - CurrentTachReadingL) * 
                           TICKS_PER_INCH;
         DistTraveledIn += (short)tmp;
         if (tmp > 2000)
            printf("CountDistanceIn 2 -- bad calc of distance %ld\n",tmp);
      }
   }
   else  // Must be right side read of tach
   {
      // Check to see if counter wrapped around
      if (NewTachReadingR < CurrentTachReadingR)
      {
         tmp = (0xFFFFFFFFUL - CurrentTachReadingR + 
                            NewTachReadingR) * TICKS_PER_INCH;
         DistTraveledIn += (short)tmp;
         if (tmp > 2000)
            printf("CountDistanceIn 3 -- bad calc of distance %ld\n",tmp);
      }
      else  // Counter not wrapped around so straight calc
      {
         tmp = (NewTachReadingR - CurrentTachReadingR) * 
                           TICKS_PER_INCH;
         DistTraveledIn += (short)tmp;
         if (tmp > 2000)
            printf("CountDistanceIn 4 -- bad calc of distance %ld\n",tmp);
      }
   }

   CurrentTachReadingR = NewTachReadingR; 
   CurrentTachReadingL = NewTachReadingL; 
   return(DistTraveledIn);
}


/****************************************************
 MoveForward -- 

   Drives  at the given speed -- Ramps the 
   velocity changes for smooth motion or bypasses
   the ramp if desired.  Also, can do turns by giving
   different speeds to left/right.
*****************************************************/
void MoveForward(short InchPerSecL, short InchPerSecR, short Ramping)
{

   if (!Ramping)   // If no ramping desired
   {
      // No ramp so just set speed
      RightWheelSpeed(InchPerSecR);
   }
   else if (InchPerSecR > (RightSpeedControl + MAX_SPD_RAMP))
   {
      RightWheelSpeed(RightSpeedControl + MAX_SPD_RAMP);
   }
   else if (InchPerSecR < (RightSpeedControl - MAX_SPD_RAMP))
   {
      RightWheelSpeed(RightSpeedControl - MAX_SPD_RAMP);
   }
   else
   {
      // Either no change or only a small change
      RightWheelSpeed(InchPerSecR);
   }
 
   if (!Ramping)   // If no ramping desired
   {
      // No ramp so just set speed
      LeftWheelSpeed(InchPerSecL);
   }
   else if (InchPerSecL > (LeftSpeedControl + MAX_SPD_RAMP))
   {
      LeftWheelSpeed(LeftSpeedControl + MAX_SPD_RAMP);
   }
   else if (InchPerSecL < (LeftSpeedControl - MAX_SPD_RAMP))
   {
      LeftWheelSpeed(LeftSpeedControl - MAX_SPD_RAMP);
   }
   else
   {
      // Either no change or only a small change
      LeftWheelSpeed(InchPerSecL);
   }
}      

/****************************************************
 WideTurn -- 

   As the name implies, this does not do a tight turn
   but rather does a sweeping turn.

  We are basing distance traveled off the inside/slow
  wheel.
  

  Returns true when turn is completed
*****************************************************/
short WideTurn(short TurnAngle)
{
   short DistOfTurn;
   short LeftRight;

   if (TurnAngle > 0)     // Always use the inside wheel
   {
      LeftRight = RIGHT;
      MoveForward(ROBOT_SPEED, ROBOT_SPEED/ROBOT_TURN_RATIO, NO_RAMPING);
   }
   else
   {
      LeftRight = LEFT;
      MoveForward(ROBOT_SPEED/ROBOT_TURN_RATIO, ROBOT_SPEED, NO_RAMPING);
   }

   DistOfTurn = (TurnAngle * TURN_CIRCUMFERANCE) / 360;

   if (CountDistance(LeftRight) >= DistOfTurn)   
      return(TRUE);   // Done when turn dist covered.

   return(FALSE);
}

/****************************************************
 DriveTowardLight -- 

   Adjusts left right speed based on light position
   in camera view.
*****************************************************/
void DriveTowardLight()
{
   short CamAngle;

// Assuming all math is done in radians -- converted to Degrees only at the end

   CamAngle = (short)(R_GetCamHorizAngle() * 180 / PI);        // Get camera angle.
   if (CamAngle > 2)  // Light to the right
      MoveForward(ROBOT_TRK_SPD,ROBOT_TRK_SPD - TRK_RATE,NO_RAMPING);   // Track right no ramp
   else if (CamAngle < -2)  // Light to the left
      MoveForward(ROBOT_TRK_SPD - TRK_RATE,ROBOT_TRK_SPD,NO_RAMPING);   // Track left no ramp
   else
      MoveForward(ROBOT_TRK_SPD,ROBOT_TRK_SPD,NO_RAMPING);   // Track straight no ramp
}

/****************************************************
 LeftWheelSpeed and RightWheelSpeed -- 

   Sets the wheel drive output to the requested
   speed.
   All ramping and special case must be handled
   above this routine -- this simply does at it is
   told.
*****************************************************/
void LeftWheelSpeed(short InchPerSec)
{
   // Convert InchPerSec to A/D reading
   // Set A/D output
   LeftSpeedControl = InchPerSec;
   R_SetLeftWheelSpeed(InchPerSec);
}

void RightWheelSpeed(short InchPerSec)
{
   // Convert InchPerSec to A/D reading
   // Set A/D output
   RightSpeedControl = InchPerSec;
   R_SetRightWheelSpeed(InchPerSec);
}

/****************************************************
 AutonomousControl -- 

   Determine entire approach to the rack
   1 - Starting anywhere in the home zone pointed basically at the center of  
       the rack. Assuming that we start near the 8 ft line of the home zone
   2 - Drive a few feet forward (to get a better angle reading of the light) 
       while allowing the camera to track the light
   3 - Assuming we want to approach straight into the dimond plate, Calculate
       the turn and distance (alignment angle and alignment dist) required to
       put us on that straight approach
   4 - Calculate the turn from the alignment move onto the approach vector and
       the distance to travel on the approach vector  (approach angle and
       approach dist)
   5 - Transition through the moves to get onto the approach vector
   6 - Travel the approach distance based on the old calculated data or as long as
       the camera can see the light, drive via the camera data up to rack
   7 - place the ring on the rack
   8 - move back away from the rack.

   NOTE -- the position of the arm and wrist will be taking place during the moves.
*****************************************************/
void AutonomousControl(void)
{

   //xxxx must start with a grab tube position
   switch (ControlState)
   {      
   case MOVE_FROM_START:
      // Move straight forward with speed up rampinp
      // and have the camera track the light
      // start moving arm to the desired position
      MoveForward(ROBOT_SPEED,ROBOT_SPEED,RAMPING);   
      Servo_Track();
//      MoveArm(SelectedArmPos);    //xxxxxxx
      // Done when we have traveled straight out a little
      if (CountDistance(LEFT) >= START_DIST)
      {
         InitDistance();         // Zero dist for next move
         CalcMoves();            // Determine path for entire movement
         ControlState = TURN_TO_ALIGN;
         // to debug comment out state change and put MoveForward(0,0,NO_RAMPING);
      }
      else                    // Don't want to waste a cycle between states
         break;               // so only break when not going to new state
   case TURN_TO_ALIGN:
      // Center the camera and turn toward the approach point  
      Home_Camera();         
//      MoveArm(SelectedArmPos);   //xxxxxxx
//      MoveGripper(SelectedArmPos);   //xxxx Move gripper based on arm position
      if (WideTurn(AlignAngle))        // Returns true when competed
      {
         InitDistance();              // Zero dist for next move
         ControlState = MOVE_TO_ALIGN;
      }
      else                    // Don't want to waste a cycle between states
         break;               // so only break when not going to new state
   case MOVE_TO_ALIGN:
      MoveForward(ROBOT_SPEED,ROBOT_SPEED,RAMPING);   // Move straight with rampinp
      Home_Camera();
//      MoveArm(SelectedArmPos);  //xxxxx
//      MoveGripper(SelectedArmPos);  // xxxxxx// Move gripper based on arm position
      if (CountDistance(LEFT) >= AlignDistIn)
      {
         InitDistance();              // Zero dist for next move
         ControlState = TURN_TO_APPROACH;
      }
      else                    // Don't want to waste a cycle between states
         break;               // so only break when not going to new state
   case TURN_TO_APPROACH:         // Note the diff between adjust and align
      Home_Camera();
//      MoveArm(SelectedArmPos);   //XXXx
//      MoveGripper(SelectedArmPos); /xxxx // Move gripper based on arm position
      if (WideTurn(ApproachAngle))     // Returns true when competed
      {
         InitDistance();              // Zero dist for next move
         DelayForCamera = 100;
         MoveForward(0,0,NO_RAMPING);   // Stop till camera tracks
         ControlState = DELAY_FOR_CAMERA;
      }
      else                    // Don't want to waste a cycle between states
         break;               // so only break when not going to new state
   case DELAY_FOR_CAMERA:         // Note the diff between adjust and align
      Servo_Track();
//      MoveArm(SelectedArmPos);   //XXXx
//      MoveGripper(SelectedArmPos); /xxxx // Move gripper based on arm position
      if (DelayForCamera <= 0)     // If done waiting for the camera
      {
         InitDistance();              // Zero dist for next move
         //xxxxx  may need to take a distance reading here via camera
         ControlState = MOVE_TO_APPROACH;
      }
      else                    // Still waiting for the camera
      {
         DelayForCamera--;
                              // Don't want to waste a cycle between states
         break;               // so only break when not going to new state
      }

   case MOVE_TO_APPROACH:              // Track light if possible
      DriveTowardLight();
      Servo_Track();
//      MoveArm(SelectedArmPos);  xxxx
//     MoveGripper(SelectedArmPos);   xxxx // Move gripper based on arm position
      if (CountDistance(LEFT) >= ApproachDistIn)  //xxx how to set this dist during approach??
      {
         InitDistance();              // Zero dist for next move
         ControlState = DROP_TUBE;
         MoveForward(0,0,NO_RAMPING);   // Stop
      }
      else                    // Don't want to waste a cycle between states
         break;               // so only break when not going to new state
   case DROP_TUBE:              
      MoveForward(0,0,NO_RAMPING);   // Stop
      if (ReleaseGripper())      // Let go
      {
         InitDistance();              // Zero dist for next move
         ControlState = ENDING_MOVE;
      }
      break;
   case ENDING_MOVE:            // Get out of there
      MoveForward(-ROBOT_SPEED,-ROBOT_SPEED,RAMPING);   // Move straight with rampinp
      break;
   }
         


}


   
/****************************************************
 CalcMoves -- 

  Unfortunatly the equations are based on a hand 
  drawing I have.  If I have time to make a text
  diagram for this comment block I will -- otherwise, 
  I am making 2 triangles - one with respect to the 
  center of the rack, the robot, and the light.  The
  other is respect to the light, the robot, and an 
  imaginary point located straight out from the light
  6 feet (number may change with testing).  The idea
  being to get to that imaginary point and face the
  rack which will make a straight approach to the 
  diamond plate.
   
*****************************************************/
void CalcMoves()
{
   short CamDist;
   short InitialTurn;
   double CamAngle;
   double Beta;
   double BetaP;
   double Lambda;
   double LambdaP;
   double AlphaP;
   double tmp;

// Assuming all math is done in radians -- converted to Degrees at the end
// Since the camera has been tracking, even if it is not exactly on, we can 
// just take where it currently is.
// I don't think we need to wait for a stable centering
   CamDist  = R_GetCamDist();
   CamAngle = R_GetCamHorizAngle();
   printf("CalcMoves - Camera Dist %d and angle %lf\n",
                                             CamDist, CamAngle*180/PI);

   // Determine Right/left based on +/- of camera angle
   If (CamAngle < 0)
   {
      // Do all the math with positive angles and then correct at the end
      InitialTurn = LEFT;
      CamAngle = -CamAngle;
   }
   else
      InitialTurn = RIGHT;

   // Calculate angle between us and rack arm at center. Beta
   Beta = asin((double)CamDist * sin(CamAngle) / (double)RACK_LEG_LEN);

   // Calculate angle between us and rack arm at light.  Lambda
   Lambda = 180 * (PI / 180) - (CamAngle + Beta);

   // Calculate angle between us and imaginary arm extended out past the light
   AlphaP = CamAngle + Beta;
   printf("Beta = %lf  Lambda = %lf  AlphaP = %lf\n",
                              Beta*180/PI, Lambda*180/PI, AlphaP*180/PI);

   // Calculate distance to a straight on approach to the light
   AlignDistIn = (short)sqrt(CamDist * CamDist + APPROACH_DIST * APPROACH_DIST - 
         (2 * CamDist * APPROACH_DIST * cos(AlphaP)));

   // Calc angle from camera angle to start of approach point
   tmp = (APPROACH_DIST * sin(AlphaP)) / AlignDistIn;
   if (tmp >= 1 || tmp <= -1)
      LambdaP = PI / 2;
   else
      LambdaP = asin(tmp);

   // Calc angle to turn once the robot is at the approach point
   tmp = CamDist * sin(AlphaP) / AlignDistIn;
   if (tmp >= 1 || tmp <= -1)
      BetaP = PI / 2;
   else
      BetaP = asin(tmp);
   printf("AlignDistIn = %d  LambdaP = %lf  BetaP = %lf\n",
                        AlignDistIn, LambdaP*180/PI, BetaP*180/PI);

   // Since I am using a sweeping turn, must adjust the distances based
   // on the distance that will be used in the turn.
   tmp = sin(CamAngle + LambdaP) * WHEEL2WHEEL_DIST + 
         sin(BetaP) * WHEEL2WHEEL_DIST;
   AlignDistIn -=  (short)tmp;

   ApproachDistIn = APPROACH_DIST - (short)tmp;

   ///////  Convert angles to deg here for integer use later on.
   // So from our current position we can calculate the alignment angle needed to
   // head to the approach vector.
   AlignAngle = (short)((CamAngle + LambdaP) * 180.0 / PI);

   ApproachAngle = (short)(-BetaP * 180.0 / PI);

   // All angles are calculated as positive -- so here, if 
   // we are turning left then we start with a neg angle
   // else if right, then need a negative angle for the
   // turn onto the approach.
   if (InitialTurn == LEFT)
   {
      AlignAngle = -AlignAngle;
   }
   else
   {
      ApproachAngle = -ApproachAngle;
   }
   printf("Aligndist/angle %d %d  -- Aproach dist/ang  %d %d\n",
                         AlignDistIn,AlignAngle,ApproachDistIn,ApproachAngle);

}


/****************************************************
 AutoInit -- 

   Initialization for autonomous mode
   
*****************************************************/
void AutoInit()
{
   InitDistance();
}

/***************************************************************

                  ARM/GRIPPER    CONTROL

****************************************************************/
void MoveArm(short Pos)
{
   //?
}
 
/***********************************************
   ReleaseGripper  -- opens to drop tube
***********************************************/
short ReleaseGripper()
{
   R_GripRelease();
   if (R_GripIsReleased())
      return(TRUE);
   return(FALSE);
}



/***************************************************************

  Robot interface functions 

***************************************************************/

/**********************************************

  R_GetTach -- reads the requested tach counter

***********************************************/
unsigned long R_GetTach(short LeftRight)
{

	if (LeftRight == LEFT)
		return(1);           // xxxxxx  return left side counter
	else
		return(2);			 // xxxxxx  return right side counter
}

/**********************************************

  R_SetLeftWheelSpeed -- Sets wheel speed
  R_SetRightWheelSpeed -- Sets wheel speed

***********************************************/

void R_SetLeftWheelSpeed(InchPerSec)
{
   // xxxxxx
}

void R_SetRightWheelSpeed(InchPerSec)
{
   //xxxxxx
}


/**********************************************

  R_GetCamDist -- Get the distance to target in inches
                  from the camera -- if this does not
                  exist then just get the up/down angle
                  and use dist = (lighheight - camera height)/tan(camera angle)
  R_GetCamHorizAngle -- Gets position of camera motor angle  -- radians
  R_GetCamVerticalAngle -- Gets position of camera motor angle -- radians
  R_GetCamHorizLightPos -- gets image position of light in camera -- calc center of light source

***********************************************/

short R_GetCamDist()
{
   // xxxxx
}


float R_GetCamHorizAngle()
{
   // xxxxxx
}

float R_GetCamVerticalAngle()
{
   // xxxxxx
}


short R_GetCamHorizLightPos()
{
   // Simple camera readings of left right in image.  xxxxxx
}


void R_GripRelease()
{
   // xxxxxx
}

short R_GripIsReleased()
{
   // xxxxxx
}

