#include <math.h>
#include "iomap.h"
#include "motion.h"
#include "control.h"
#include "encoder.h"
#include "camera.h"
#include "tracking.h"

/* START: Tweakable parameters                       */

// Distance, in ticks, to drive out at start before using
// camera tracking.
#define START_DIST (TICKS_PER_FOOT * 1L)

// Distance, in ticks, we want to approach the rack from
#define APPROACH_DIST (TICKS_PER_FOOT * 5L)

// Length, in ticks, of rack leg from center of rack
#define RACK_LEG_LEN (TICKS_PER_FOOT * 4L)

/* END: Tweakable parameters                         */

static enum
{
    MOVE_FROM_START,
    CAMERA_SEARCH,
    TURN_TO_ALIGN,       
    MOVE_TO_ALIGN,
    TURN_TO_APPROACH,        
    MOVE_TO_APPROACH,       
    DROP_TUBE,              
    ENDING_MOVE, 
    END_OF_STATE_TABLE,
    COMPLETED
} ControlState;

static int state_initialized;

static PIDData left;
static PIDData right;


/* START: Outputs of CalcMoves()                   */
/* All in # of ticks, usable by motion.h functions */

static long align_turn;     // Turn amount from start to
                            //  align vector
static long align_dist;     // Distance from start to
                            //  align/approach intersection
static long approach_turn;  // Turn amount from align 
                            //  vector to approach vector
static long approach_dist;  // Distance from align/approach
                            //  intersection to rack

/* END: Outputs of CalcMoves()                     */


/* Input parameters based on strategy selection    */

static enum                 // Look more to left or right
{                           //  when 2 lights are in sight
    WEIGHTED_LEFT,
    WEIGHTED_RIGHT
} WeightDirection;

static ArmCommand score_height;



void camscore_init(ArmCommand sh, int weight_left)
{
    score_height = sh;
    WeightDirection = weight_left ? WEIGHTED_LEFT : WEIGHTED_RIGHT;
    state_initialized = 0;
    ControlState = MOVE_FROM_START;

    Init_Drive_PIDData_Struct(&left, &pwm_left_drive);
    Init_Drive_PIDData_Struct(&right, &pwm_right_drive);
}




/****************************************************
Main camera scoring processing function

   Determine entire approach to the rack
   1 - Starting anywhere in the home zone pointed basically at the center of  
       the rack. Assuming that we start near the 8 ft line of the home zone
   2 - Drive a few feet forward (to get a better angle reading of the light) 
       while allowing the camera to track the light
   3 - Assuming we want to approach straight into the dimond plate, Calculate
       the turn and distance (alignment angle and alignment dist) required to
       put us on that straight approach
   4 - Calculate the turn from the alignment move onto the approach vector and
       the distance to travel on the approach vector  (approach angle and
       approach dist)
   5 - Transition through the moves to get onto the approach vector
   6 - Travel the approach distance based on the old calculated data or as long as
       the camera can see the light, drive via the camera data up to rack
   7 - place the ring on the rack
   8 - move back away from the rack.

   NOTE -- the position of the arm and wrist will be taking place during the moves.
*****************************************************/

void AutonomousControl()
{
    switch (ControlState) {
      case MOVE_FROM_START:
        if (!state_initialized) {
            state_initialized = 1;
            Reset_Left_Encoder_Count();
            Reset_Right_Encoder_Count();
            Set_PIDData_Target(&left, START_DIST);
            Set_PIDData_Target(&right, START_DIST);
        }
        Arm_Process(127, 127, WRIST_SCORELO);
        if (!Drive_Ticks(&left, &right))
            break;

        pwm_wrist = 127;
        ControlState = CAMERA_SEARCH;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case CAMERA_SEARCH:
        // XXX - wait for confidence good on multiple pkts?
        if (GT_Packet_Data.confidence < CONFIDENCE_THRESHOLD_DEFAULT)
            break;

        // XXX - figure out all the movement stuff
        // CalcMoves()
        ControlState = TURN_TO_ALIGN;
        // NOTE: Falls through!!!

      case TURN_TO_ALIGN:
        if (!state_initialized) {
            state_initialized = 1;
            Reset_Left_Encoder_Count();
            Reset_Right_Encoder_Count();
            Init_Drive_Turn(&left, &right, align_turn);
        }

        if (!Turn_Ticks(&left, &right))
            break;

        ControlState = MOVE_TO_ALIGN;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case MOVE_TO_ALIGN:
        if (!state_initialized) {
            state_initialized = 1;
            Reset_Left_Encoder_Count();
            Reset_Right_Encoder_Count();
            Set_PIDData_Target(&left, align_dist);
            Set_PIDData_Target(&right, align_dist);
        }

        if (!Drive_Ticks(&left, &right))
            break;

        ControlState = TURN_TO_APPROACH;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case TURN_TO_APPROACH:
        if (!state_initialized) {
            state_initialized = 1;
            Reset_Left_Encoder_Count();
            Reset_Right_Encoder_Count();
            Init_Drive_Turn(&left, &right, approach_turn);
        }

        if (!Turn_Ticks(&left, &right))
            break;

        ControlState = MOVE_TO_APPROACH;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case MOVE_TO_APPROACH:
        // XXX - Move arm to desired scoring position
        // XXX - Track light and adjust drive if possible
        //       or.... recalc a drive angle/distance?
        // XXX - drive at most approach_dist

        ControlState = DROP_TUBE;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case DROP_TUBE:
        // XXX - Just open gripper or drop arm a bit then
        //       open gripper??

        ControlState = ENDING_MOVE;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case ENDING_MOVE:            // Get out of there
        if (!state_initialized) {
            state_initialized = 1;
            Reset_Left_Encoder_Count();
            Reset_Right_Encoder_Count();
            Set_PIDData_Target(&left, -(approach_dist / 2L));
            Set_PIDData_Target(&right, -(approach_dist / 2L));
        }
        // XXX - Put arm to park???

        if (!Drive_Ticks(&left, &right))
            break;

        ControlState = COMPLETED;
        state_initialized = 0;
        // NOTE: Falls through!!!!

      case COMPLETED:
        break;
   }

}



#define PI (3.14159f)

void CalcMoves() {
    int CamDist;
    double CamAngle;
    double Beta;
    double BetaP;
    double Lambda;
    double LambdaP;
    double AlphaP;
    double tmp;

    // Assuming all math is done in radians -- converted to Degrees only at the end
    // XXX - need to provide these
    CamDist  = R_GetCamDist();
    CamAngle = R_GetCamHorizAngle();

    // Calculate angle between us and rack arm at center. Beta
    Beta = asin((float)CamDist * sin(CamAngle) / (float)RACK_LEG_LEN);

    // Calculate angle between us and rack arm at light.  Lambda
    Lambda = 180 * (PI / 180) - (CamAngle + Beta);

    // Calculate angle between us and imaginary arm extended out past the light
    AlphaP = CamAngle + Beta;

    // Calculate distance to a straight on approach to the light
    align_dist = (long)
         sqrt(CamDist * CamDist + APPROACH_DIST * APPROACH_DIST - 
         (2 * CamDist * APPROACH_DIST * cos(AlphaP)));

    // Calc angle from camera angle to start of approach point
    tmp = (APPROACH_DIST * sin(AlphaP)) / align_dist;
    if (tmp >= 1 || tmp <= -1)
       LambdaP = PI / 2;
    else
       LambdaP = asin(tmp);

    // Calc angle to turn once the robot is at the approach point
    tmp = CamDist * sin(AlphaP) / align_dist;
    if (tmp >= 1 || tmp <= -1)
       BetaP = PI / 2;
    else
       BetaP = asin(tmp);


    ///////  Convert angles to deg here for integer use later on.
    // So from our current position we can calculate the alignment angle needed to
    // head to the approach vector.
    // XXX - these are wrong -- need conversion to ticks
    align_turn = (long)((CamAngle + LambdaP) * 180.0 / PI);

    align_turn = (long)(-BetaP * 180.0 / PI);

    approach_dist = APPROACH_DIST;
}





