/*******************************************************************************
*
*	TITLE:		control.c 
*
*	VERSION:	0.1                           
*
*	DATE:		02-03-2007
*
*	AUTHOR:		Team 1511
*
*	DESCRIPTION:	This file contains controls for the arm, manipulator, and turret
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	03-Feb-2007  0.2  Cannibalized file for arm/manip/turret control 
* 	24-Jan-2007  0.1  Curse you Kevin.org!
*
*******************************************************************************/

#include <stdio.h>
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "user_routines.h"
#include "serial_ports.h"
#include "encoder.h"
#include "iomap.h"
#include "motion.h"
#include "control.h"


// Uncomment if you want detailed printouts regarding what PIDs are doing
//#define DEBUG_ARM

static void Init_Turret_PIDData_Struct(PIDData* control, unsigned char* pwm);
static void Init_Arm_PIDData_Struct(PIDData* control, unsigned char* pwm);
static void Init_Wrist_PIDData_Struct(PIDData* control, unsigned char* pwm);

static void Arm_Read_Sensors(void);
static void Set_Turret_PWM(unsigned char new);
static void Set_Arm_PWM(unsigned char new);
static int Set_Wrist_Angle(int angle);
static void Auto_Wrist(void);
static void Manual_Wrist(void);


/* 
 * True if the wrist is within a safe tolerance (WRIST_DELTA)
 * of the arm height-based minimum amount. If false,
 * we're way out of our 72x72 box
 */
static int Wrist_Safe(void);




static const rom unsigned char turret_drive_lut[];
static const rom unsigned char arm_drive_lut[];
static const rom unsigned char auto_wrist_lut[];

#define AUTO_WRIST_LUT_LEN 550


static int arm_pos;
static int wrist_pos;
static int turret_pos;
static int wrist_min;

static char broken_turret;
static char broken_wrist;
static char broken_arm;
static char broken_manip;

static PIDData TurretPID;
static PIDData ArmPID;
static PIDData WristPID;
static int wrist_offset;


typedef struct {
  int turret_used;
  // Final target
  int wrist_tgt;
  int wrist_tgt_set;
} ArmAutoState;
static ArmAutoState arm_auto_state;




void Arm_Init(void)
{
	Init_Turret_PIDData_Struct(&TurretPID, &pwm_turret);
	Init_Arm_PIDData_Struct(&ArmPID, &pwm_arm);
	Init_Wrist_PIDData_Struct(&WristPID, &pwm_wrist);
  Arm_Reset();
  broken_turret = 1;
  broken_arm = 1;
  broken_wrist = 1;
  broken_manip = 1;
}

void Arm_Reset(void)
{
	wrist_offset = 0;
}

void Arm_Sensor_Set_Broken(char turret, char arm, char wrist, char manip)
{
  broken_turret = turret;
  broken_arm = arm;
  broken_wrist = wrist;
  broken_manip = manip;
}

static void Init_Turret_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  Init_PIDData_Struct(control, pwm);
  control->tolerance = 2;
  control->kp_num = 3;
  control->kp_den = 2;
  control->ki_num = 0;
  control->ki_den = 1;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 127;
  control->i_out_max = 0;
  control->output_min = 20;
  control->output_max = 127;
}

static void Init_Arm_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  Init_PIDData_Struct(control, pwm);
  control->tolerance = 5;
  control->kp_num = -1;
  control->kp_den = 1;
  control->ki_num = 0;
  control->ki_den = 1;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 35;
  control->i_out_max = 0;
  control->output_min = 20;
  control->output_max = 35;
}

static void Init_Wrist_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  Init_PIDData_Struct(control, pwm);
  control->tolerance = 10;
  control->kp_num = -3;
  control->kp_den = 1;
  control->ki_num = 0;
  control->ki_den = 1;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 127;
  control->i_out_max = 0;
  control->output_min = 25;
  control->output_max = 127;
}

static void Arm_Read_Sensors(void)
{
  int diff;

  arm_pos = Get_Analog_Value(anain_arm);
  wrist_pos = Get_Analog_Value(anain_wrist);
  turret_pos = Get_Analog_Value(anain_turret);

  diff = arm_pos - ARM_MIN;
  if (diff < 0)
    diff = 0;
  else if (diff >= AUTO_WRIST_LUT_LEN)
    diff = AUTO_WRIST_LUT_LEN - 1;

  wrist_min = auto_wrist_lut[diff] + WRIST_MIN;
}

void Arm_Process(unsigned char arm, int turret, int wrist)
{
  static int i = 0;

  Arm_Read_Sensors();

#ifdef DEBUG_ARM
  if (i > 30) {
    printf("Arm in = %d\r\n", (int)arm);
    printf("Wrist in = %d\r\n", (int)wrist);
    printf("Turret in = %d\r\n", (int)turret);
    printf("Arm Pos = %d\r\n", arm_pos);
    printf("Wrist Pos = %d\r\n", wrist_pos);
    printf("Turret Pos = %d\r\n\r\n", turret_pos);
  }
#endif

  // Arm movement
  Set_Arm_PWM(arm);


  // Wrist movement
  if (broken_wrist) {
    Manual_Wrist();

  } else {

    if (wrist < 0)
      // Auto wrist
      Auto_Wrist();
    else
      Set_Wrist_Angle(wrist);
  }

  // Turret movement
  if ((turret == -1) && !broken_turret)   // Center turret
  {
    if (arm_pos > ARM_MIN_ROTATE)
    {
      Set_PIDData_Target(&TurretPID, TURRET_CENTER);
      PID_Control(&TurretPID, turret_pos);
    }
    else
    {
      Set_Turret_PWM(127);
    }
  }
  else    // Manual turret drive
  {
    Set_Turret_PWM(turret);
  }

#ifdef DEBUG_ARM
  if (i++ > 30) {
    i = 0;
    printf("Arm PWM = %d\r\n", pwm_arm);
    printf("Wrist PWM = %d\r\n", pwm_wrist);
    printf("Turret PWM = %d\r\n\r\n", pwm_turret);
  }
#endif
}






static void Set_Arm_PWM(unsigned char new)
{
  new = arm_drive_lut[new];

  // NOTE: for arm drives, < 127 raises arm
  if (!broken_arm)
  {
    if (new == 127) {
      pwm_arm = 127;
      return;
    }

    if (new < 127)
    {
      if (arm_pos >= ARM_MAX)
      {
        pwm_arm = 127;
        return;
      }
      if (!broken_wrist && !Wrist_Safe())
      {
        pwm_arm = (unsigned char)(((int)new - 127) / 2 + 127);
        return;
      }
    }
    else if (new > 127)
    {
      int min;
      if (!broken_turret && !Turret_Centered())
        min = ARM_MIN_ROTATE;
      else
        min = ARM_MIN;
      if (arm_pos <= min)
      {
        pwm_arm = 127;
        return;
      }
      if (!broken_wrist && !Wrist_Safe())
      {
        pwm_arm = (unsigned char)(((int)new - 127) / 2 + 127);
        return;
      }
    }
  }
  pwm_arm = new;
}



static void Set_Turret_PWM(unsigned char new)
{
  unsigned char centered;

  new = turret_drive_lut[new];

  if (!broken_turret)
  {

    if (new != 127)
    {
      int centered = Turret_Centered();

      if (!broken_arm && centered && arm_pos <= ARM_MIN_ROTATE)
      {
        new = 127;
      }
      else if (new > 127 && turret_pos >= TURRET_MAX)
      {
        new = 127;
      }
      else if (new < 127 && turret_pos <= TURRET_MIN)
      {
        new = 127;
      }
    }

  }

  pwm_turret = new;

}




// Controls for when wrist sensor is broken
static void Manual_Wrist(void)
{
  if (oi_wrist_fine_up)
    pwm_wrist = WRIST_MANUAL_UP;
  else if (oi_wrist_fine_down)
    pwm_wrist = WRIST_MANUAL_DOWN;
  else
    pwm_wrist = 127;
}


static void Auto_Wrist(void)
{
  if (oi_wrist_fine_down) {
    ++wrist_offset;
    // Overflow prevention
    if (wrist_offset > WRIST_FINE_MAX)
      wrist_offset = WRIST_FINE_MAX;
  } else if (oi_wrist_fine_up) {
    wrist_offset -= 2;
    // Underflow prevention
    if (wrist_offset < 0)
      wrist_offset = 0;
  }

  Set_Wrist_Angle(wrist_min + wrist_offset);
}


static int Set_Wrist_Angle(int angle)
{
  if (angle < WRIST_MIN)
    angle = WRIST_MIN;
  else if (angle > WRIST_MAX)
    angle = WRIST_MAX;

  if (!broken_arm)
  {
    if (angle >= wrist_min)
      Set_PIDData_Target(&WristPID, angle);
    else
      Set_PIDData_Target(&WristPID, wrist_min);
	} 
	else 
	{
    Set_PIDData_Target(&WristPID, angle);
	}

  if (!PID_Control(&WristPID, (long)wrist_pos)) {

    // Extent check: Override the PID if it has run us 
    // too far and wants us to go still farther
    if ((pwm_wrist < 127 && wrist_pos >= WRIST_MAX) ||
        (pwm_wrist > 127 && wrist_pos <= WRIST_MIN))
      pwm_wrist = 127;
    return 0;
  }
  return 1;

}



int Wrist_At_Limit(void)
{
  int min = !broken_arm ? wrist_min : WRIST_MIN;
  return wrist_pos <= min || wrist_pos >= WRIST_MAX;
}

int Arm_At_Limit(void)
{
  int min = !broken_turret && !Turret_Centered() ?
                ARM_MIN_ROTATE : ARM_MIN;

  return arm_pos >= ARM_MAX || arm_pos <= min ||
          !broken_wrist && !Wrist_Safe();
}

int Turret_At_Limit(void)
{
  int min, max;
  if (!broken_arm && arm_pos <= ARM_MIN_ROTATE) {
      max = TURRET_CENTER + TURRET_CENTER_TOLERANCE;
      min = TURRET_CENTER - TURRET_CENTER_TOLERANCE;
  } else {
      max = TURRET_MAX;
      min = TURRET_MIN;
  }
  return turret_pos <= TURRET_MIN || 
         turret_pos >= TURRET_MAX;
}


int Turret_Centered(void)
{
  return ((turret_pos < TURRET_CENTER + TURRET_CENTER_TOLERANCE) && (turret_pos > TURRET_CENTER - TURRET_CENTER_TOLERANCE));
}

int Wrist_Safe(void)
{
  int d = wrist_pos - wrist_min;
  return !(d < -WRIST_DELTA);
}

#if 0

#define MANIP_BAD_NONE 0
#define MANIP_BAD_OPEN 1
#define MANIP_BAD_CLOSE 2

// We start closed.....
static int last_manip_closed = 1;
static int manip_dir_bad = MANIP_BAD_CLOSE;
static unsigned char last_pwm_manip = MANIP_DRIVE_CLOSE;

// Based on current values in control.h for MANIP_DRIVE_OPEN/CLOSED
static int Manip_Bad_From_Drive_Dir(unsigned char val)
{
  if (val > 127)
    return MANIP_BAD_CLOSE;
  else if (val < 127)
    return MANIP_BAD_OPEN;
  else
    return MANIP_BAD_NONE;
}

static int Manip_Check_Set(unsigned char newval)
{
printf("Manip old %u new %u last %u cur %u\r\n",
        (unsigned int) pwm_manip, (unsigned int)newval,
         (unsigned int)last_manip_closed, 
          (unsigned int)digin_manip_limit_close);
printf("State: %d\r\n", manip_dir_bad);
  if (!broken_manip) {
    if (digin_manip_limit_close && !last_manip_closed) {
      // We just transitioned to closed - don't allow
      // motion in the direction that got us here anymore
      manip_dir_bad = Manip_Bad_From_Drive_Dir(last_pwm_manip);
    } else if (!digin_manip_limit_close) {
      manip_dir_bad = MANIP_BAD_NONE;
    }
  
    last_manip_closed = digin_manip_limit_close;
  
    if (manip_dir_bad != MANIP_BAD_NONE &&
            manip_dir_bad == Manip_Bad_From_Drive_Dir(newval)) {
      pwm_manip = 127;
      return 1;
    }
  } else {
    last_manip_closed = 0;
    manip_dir_bad = MANIP_BAD_NONE;
  }
  if (pwm_manip != 127)
    last_pwm_manip = pwm_manip;
  pwm_manip = newval;
  printf("New manip: %u\r\n", (unsigned int)pwm_manip);
  return 0;
}

// Returns 1 for done moving, 0 for still at it
int Open_Manip(void)
{
  if (Get_Manip_State() == 0) {
    Manip_Check_Set(127);
    return 1;
  } else {
    Manip_Check_Set(MANIP_DRIVE_OPEN);
    return 0;
  }
}

void Stop_Manip(void)
{
  Manip_Check_Set(127);
}

// Returns 1 for done moving, 0 for still at it
int Close_Manip(void)
{
  return Manip_Check_Set(MANIP_DRIVE_CLOSE);
}

#else


// Returns 1 for done moving, 0 for still at it
int Open_Manip(void)
{
  if (Get_Manip_State() == 0) {
    pwm_manip = 127;
    return 1;
  } else {
    pwm_manip = MANIP_DRIVE_OPEN;
    return 0;
  }
}

// Returns 1 for done moving, 0 for still at it
int Close_Manip(void)
{
  int x = Get_Manip_State();
  if (x == 1 || x == 2) {
    pwm_manip = 127;
    return 1;
  } else {
    pwm_manip = MANIP_DRIVE_CLOSE;
    return 0;
  }
}


void Stop_Manip(void)
{
  pwm_manip = 127;
}
#endif

// Reads data from manipulator sensors and limit switches to determine state.
// Returns 0 for open, 1 for closed, and 3 for in between.
int Get_Manip_State(void)
{
  int ret = 3;

  if (!broken_manip) {
    if (digin_manip_limit_open)
      ret = 0;
    else if (digin_manip_limit_close)
      ret = 1;
  }
  
  return ret;
}



void Arm_Auto_Init(ArmCommand command)
{
  if (command == ARM_CMD_PARK) {
    arm_auto_state.turret_used = 1;
    Set_PIDData_Target(&TurretPID, TURRET_CENTER);
    Set_PIDData_Target(&ArmPID, ARM_MIN);
    arm_auto_state.wrist_tgt = WRIST_PARK;
  
  } else if (command == ARM_CMD_PICKUP) {
    arm_auto_state.turret_used = 1;
    Set_PIDData_Target(&TurretPID, TURRET_CENTER);
    Set_PIDData_Target(&ArmPID, ARM_PICKUP);
    arm_auto_state.wrist_tgt = WRIST_PICKUP;
  
  } else if (command == ARM_CMD_LOW) {
    arm_auto_state.turret_used = 1;
    Set_PIDData_Target(&TurretPID, TURRET_CENTER);
    Set_PIDData_Target(&ArmPID, ARM_SCORELO);
    arm_auto_state.wrist_tgt = WRIST_SCORELO;

  } else if (command == ARM_CMD_MID) {
    arm_auto_state.turret_used = 0;
    Set_PIDData_Target(&ArmPID, ARM_SCOREMID);
    arm_auto_state.wrist_tgt = WRIST_SCOREMID;

  } else if (command == ARM_CMD_HI){
    arm_auto_state.turret_used = 0;
    Set_PIDData_Target(&ArmPID, ARM_SCOREHI);
    arm_auto_state.wrist_tgt = WRIST_SCOREHI;
  }

  wrist_offset = 0;
  arm_auto_state.wrist_tgt_set = 1;
  Set_PIDData_Target(&WristPID, arm_auto_state.wrist_tgt);
}


unsigned char Arm_Auto_Process(void)
{
  int wrist_done;

  Arm_Read_Sensors();

  /* Check for required sensor availability.
   * Turret sensor not required since it is so prone
   *   to breakage and it only serves to prevent hitting
   *   the chassis
   */
  if (broken_arm || broken_wrist) {
    // We can't deal....
    pwm_arm = pwm_turret = pwm_wrist = 127;
    return 0;
  }

  // Process turret first, if needed and able
  // See above for why we continue anyway if turret broken
  if (!broken_turret && arm_auto_state.turret_used) {
    
    if (!PID_Control(&TurretPID, turret_pos)) {
      // Check turret extents
      if (pwm_turret < 127 && turret_pos <= TURRET_MIN ||
          pwm_turret > 127 && turret_pos >= TURRET_MAX)
        pwm_turret = 127;

      pwm_arm = pwm_wrist = 127;      
      return 0;
    }
  } else {

    pwm_turret = 127;
  }


  // Check wrist against minimum, which is computed from 
  // arm height.
  if (arm_auto_state.wrist_tgt >= wrist_min) {
    // Wrist ok -- set final destination tgt if not
    // already finished.
    if (!arm_auto_state.wrist_tgt_set) {
      arm_auto_state.wrist_tgt_set = 1;
      Set_PIDData_Target(&WristPID, arm_auto_state.wrist_tgt);
    }
  } else {
    // Wrist in a bad place -- target the minimum
    arm_auto_state.wrist_tgt_set = 0;
    Set_PIDData_Target(&WristPID, wrist_min);
  }

  // Process wrist
  if (!PID_Control(&WristPID, (long)wrist_pos)) {
    // Extent check
    if (pwm_wrist < 127 && wrist_pos >= WRIST_MAX ||
        pwm_wrist > 127 && wrist_pos <= WRIST_MIN)
      pwm_wrist = 127;

    wrist_done = 0;
  } else {
    wrist_done = arm_auto_state.wrist_tgt_set;
  }

  if (!Wrist_Safe()) {  
    pwm_arm = 127;
  } else {
    // Extent check after PID, dependant on motor direction
    // NOTE: for arm drives, < 127 raises arm
    if (!PID_Control(&ArmPID, arm_pos)) {  
      if ((pwm_arm > 127 && arm_pos <= ARM_MIN) ||
          (pwm_arm < 127 && arm_pos >= ARM_MAX))
        pwm_arm = 127;
      return 0;
    }
    return (unsigned char)wrist_done;
  }
  return 0;
}








//f(x)=(x^3) * (64/2048383) + 127
static const rom unsigned char turret_drive_lut [] = {
	0x3f, 0x40, 0x42, 0x43, 0x45, 0x46, 0x48, 0x49, 
	0x4a, 0x4c, 0x4d, 0x4e, 0x4f, 0x51, 0x52, 0x53, 
	0x54, 0x55, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 
	0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x62, 0x63, 
	0x64, 0x65, 0x66, 0x67, 0x67, 0x68, 0x69, 0x6a, 
	0x6a, 0x6b, 0x6c, 0x6c, 0x6d, 0x6e, 0x6e, 0x6f, 
	0x70, 0x70, 0x71, 0x71, 0x72, 0x72, 0x73, 0x73, 
	0x74, 0x74, 0x75, 0x75, 0x76, 0x76, 0x76, 0x77, 
	0x77, 0x78, 0x78, 0x78, 0x79, 0x79, 0x79, 0x7a, 
	0x7a, 0x7a, 0x7a, 0x7b, 0x7b, 0x7b, 0x7b, 0x7c, 
	0x7c, 0x7c, 0x7c, 0x7c, 0x7d, 0x7d, 0x7d, 0x7d, 
	0x7d, 0x7d, 0x7d, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 
	0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7f, 0x7f, 
	0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
	0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
	0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
	0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
	0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
	0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
	0x7f, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
	0x80, 0x80, 0x80, 0x80, 0x81, 0x81, 0x81, 0x81, 
	0x81, 0x81, 0x81, 0x82, 0x82, 0x82, 0x82, 0x82, 
	0x83, 0x83, 0x83, 0x83, 0x84, 0x84, 0x84, 0x84, 
	0x85, 0x85, 0x85, 0x86, 0x86, 0x86, 0x87, 0x87, 
	0x88, 0x88, 0x88, 0x89, 0x89, 0x8a, 0x8a, 0x8b, 
	0x8b, 0x8c, 0x8c, 0x8d, 0x8d, 0x8e, 0x8e, 0x8f, 
	0x90, 0x90, 0x91, 0x92, 0x92, 0x93, 0x94, 0x94, 
	0x95, 0x96, 0x97, 0x97, 0x98, 0x99, 0x9a, 0x9b, 
	0x9c, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 
	0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa9, 0xaa, 0xab, 
	0xac, 0xad, 0xaf, 0xb0, 0xb1, 0xb2, 0xb4, 0xb5, 
	0xb6, 0xb8, 0xb9, 0xbb, 0xbc, 0xbe, 0xbf, 0xbf
};

// Logitech Y axis only goes from about 6 to about 240
// f(x) = x * (80 / 127) + 127 for 16 rows
// f(x) = x * (30 / 127) + 127 for 16 rows
static const rom unsigned char arm_drive_lut [] = {
	47, 48, 48, 49, 50, 50, 51, 51, 
	52, 53, 53, 54, 55, 55, 56, 56, 
	57, 58, 58, 59, 60, 60, 61, 61, 
	62, 63, 63, 64, 65, 65, 66, 67, 
	67, 68, 68, 69, 70, 70, 71, 72, 
	72, 73, 73, 74, 75, 75, 76, 77, 
	77, 78, 78, 79, 80, 80, 81, 82, 
	82, 83, 84, 84, 85, 85, 86, 87, 
	87, 88, 89, 89, 90, 90, 91, 92, 
	92, 93, 94, 94, 95, 96, 96, 97, 
	97, 98, 99, 99, 100, 101, 101, 102, 
	102, 103, 104, 104, 105, 106, 106, 107, 
	107, 108, 109, 109, 110, 111, 111, 112, 
	113, 113, 114, 114, 115, 116, 116, 117, 
	118, 118, 119, 119, 120, 121, 121, 122, 
	123, 123, 124, 124, 125, 126, 126, 127, 
	127, 127, 128, 128, 128, 128, 129, 129, 
	129, 129, 130, 130, 130, 130, 131, 131, 
	131, 131, 131, 132, 132, 132, 132, 133, 
	133, 133, 133, 134, 134, 134, 134, 135, 
	135, 135, 135, 136, 136, 136, 136, 136, 
	137, 137, 137, 137, 138, 138, 138, 138, 
	139, 139, 139, 139, 140, 140, 140, 140, 
	140, 141, 141, 141, 141, 142, 142, 142, 
	142, 143, 143, 143, 143, 144, 144, 144, 
	144, 144, 145, 145, 145, 145, 146, 146, 
	146, 146, 147, 147, 147, 147, 148, 148, 
	148, 148, 148, 149, 149, 149, 149, 150, 
	150, 150, 150, 151, 151, 151, 151, 152, 
	152, 152, 152, 153, 153, 153, 153, 153, 
	154, 154, 154, 154, 155, 155, 155, 155, 
	156, 156, 156, 156, 157, 157, 157, 157
};


static rom const unsigned char auto_wrist_lut[] = 
{
  25, 25, 25, 25, 25, 25, 25, 26, 
  26, 27, 27, 28, 28, 29, 29, 30, 
  30, 30, 31, 31, 32, 32, 33, 33, 
  34, 34, 35, 35, 36, 36, 36, 37, 
  37, 38, 38, 39, 39, 40, 40, 41, 
  41, 41, 42, 42, 43, 43, 44, 44, 
  45, 45, 46, 46, 47, 47, 47, 48, 
  48, 49, 49, 50, 50, 51, 51, 52, 
  52, 52, 53, 53, 54, 54, 55, 55, 
  56, 56, 57, 57, 58, 58, 58, 59, 
  59, 60, 60, 61, 61, 62, 62, 63, 
  63, 63, 64, 64, 65, 65, 66, 66, 
  67, 67, 68, 68, 68, 69, 69, 70, 
  70, 71, 71, 72, 72, 73, 73, 74, 
  74, 74, 75, 75, 76, 76, 77, 77, 
  78, 78, 79, 79, 79, 80, 80, 81, 
  81, 82, 82, 83, 83, 84, 84, 85, 
  85, 85, 86, 86, 87, 87, 88, 88, 
  89, 89, 90, 90, 90, 90, 89, 89, 
  89, 89, 88, 88, 88, 88, 87, 87, 
  87, 87, 86, 86, 86, 86, 86, 85, 
  85, 85, 85, 84, 84, 84, 84, 83, 
  83, 83, 83, 82, 82, 82, 82, 82, 
  81, 81, 81, 81, 80, 80, 80, 80, 
  79, 79, 79, 79, 79, 78, 78, 78, 
  78, 77, 77, 77, 77, 76, 76, 76, 
  76, 75, 75, 75, 75, 75, 74, 74, 
  74, 74, 73, 73, 73, 73, 72, 72, 
  72, 72, 71, 71, 71, 71, 71, 70, 
  70, 70, 70, 69, 69, 69, 69, 68, 
  68, 68, 68, 67, 67, 67, 67, 66, 
  66, 66, 65, 65, 65, 64, 64, 63, 
  63, 63, 62, 62, 62, 61, 61, 61, 
  60, 60, 60, 59, 59, 59, 58, 58, 
  57, 57, 57, 56, 56, 56, 55, 55, 
  55, 54, 54, 54, 53, 53, 53, 52, 
  52, 51, 51, 51, 50, 50, 50, 49, 
  49, 49, 48, 48, 48, 48, 47, 47, 
  47, 47, 46, 46, 46, 46, 45, 45, 
  45, 44, 44, 44, 44, 43, 43, 43, 
  43, 42, 42, 42, 42, 41, 41, 41, 
  41, 40, 40, 40, 39, 39, 39, 39, 
  38, 38, 38, 38, 37, 37, 37, 37, 
  36, 36, 36, 35, 35, 35, 35, 34, 
  34, 34, 34, 33, 33, 33, 33, 32, 
  32, 32, 32, 31, 31, 31, 30, 30, 
  30, 30, 29, 29, 29, 29, 28, 28, 
  28, 28, 27, 27, 27, 26, 26, 26, 
  26, 25, 25, 25, 25, 24, 24, 24, 
  24, 23, 23, 23, 23, 22, 22, 22, 
  21, 21, 21, 21, 20, 20, 20, 20, 
  19, 19, 19, 19, 18, 18, 18, 17, 
  17, 17, 17, 16, 16, 16, 16, 15, 
  15, 15, 15, 14, 14, 14, 14, 13, 
  13, 13, 12, 12, 12, 12, 11, 11, 
  11, 11, 10, 10, 10, 10, 9, 9, 
  9, 8, 8, 8, 8, 7, 7, 7, 
  7, 6, 6, 6, 6, 5, 5, 5, 
  5, 4, 4, 4, 3, 3, 3, 3, 
  2, 2, 2, 2, 1, 1, 1, 1, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0
};
