#include <stdio.h>
#include "iomap.h"
#include "motion.h"
#include "control.h"
#include "encoder.h"
#include "camera.h"
#include "tracking.h"
#include "lightseek.h"


/* How many consecutive good cam packets constitutes locking on? */
#define CONFIDENCE_COUNT  5

#define TWO_LIGHT_WIDTH 50

/* Max pwm value for drive train while looking for light */
#define SEARCH_DRIVE_PWM_MAX 40

// Base pwm drive value for guided driving
// Also used for base drive going straight
#define BASE_DRIVE_LEFT 32
#define BASE_DRIVE_RIGHT 32

// Max amount of drive differential to adjust for pan
#define PAN_DRIVE_ADJ_MAX 25


static enum
{
    PICKUP_BACKUP,
    PICKUP_DROPARM,
    PICKUP_GRAB,
    INITIAL_SEARCH,
    DRIVE_CURVE_SEARCH,
    DRIVE_STRAIGHT_SEARCH,
    GUIDED_DRIVE,
    ARM_TO_SCORE,
    DRIVE_TO_SCORE,
    RELEASE_TUBE,
    BACK_OFF,
    COMPLETED
} LightSeekState;


/* Input parameters based on strategy selection    */
static enum                 // Look more to left or right
{                           //  when 0 or 2 lights are in sight
    WEIGHTED_LEFT,
    WEIGHTED_RIGHT
} WeightDirection;


static void setMaxDistFromTilt(void);
static int isAtApproachDist(void);
static void setDriveFromPan(void);
static void setDriveStraight(void);



static long approach_dist;

/* Arm Command for the user-configured leg to score */
static ArmCommand score_height;
static ArmCommand release_height;

static int state_initialized;
static long max_search_dist;
static int pickup_stop;


void LightSeek_Init(ArmCommand sh, int weight_left, int pickup_mode)
{
    score_height = sh;
    if (sh == ARM_CMD_HI)
        release_height = ARM_CMD_MID;
    else if (sh == ARM_CMD_MID)
        release_height = ARM_CMD_LOW;
    else
        release_height = ARM_CMD_PICKUP;
    WeightDirection = weight_left ? WEIGHTED_LEFT : WEIGHTED_RIGHT;
    state_initialized = 0;
    max_search_dist = 2L * TICKS_PER_FOOT;    // Safety....
    approach_dist = 16L * TICKS_PER_INCH;

    switch (pickup_mode) {
        case PICKUP_ONLY:
            LightSeekState = PICKUP_BACKUP;
            pickup_stop = 1;
            break;
        case PICKUP_SKIP:
            LightSeekState = INITIAL_SEARCH;
            pickup_stop = 0;
            break;
        case PICKUP_COMPLETE:
        default:
            LightSeekState = PICKUP_BACKUP;
            pickup_stop = 0;
            break;
    }

}



void LightSeek_Process(void)
{
start_over:
    printf("State: %d\r\n", LightSeekState);

    switch(LightSeekState) {
      case PICKUP_BACKUP:
      {
        unsigned char arm;
        int drive;
        int manip;

        if (!state_initialized) {
            state_initialized = 1;
            Arm_Auto_Init(ARM_CMD_PICKUP);
            Init_Drive_Straight(-18L * TICKS_PER_INCH);
        }
        state_initialized++;

        // We need to wait for the arm & drive
        arm = Arm_Auto_Process();
        drive = Drive_Ticks();
        manip = Open_Manip() || state_initialized > 60;

        if (!arm || !drive)
            break;
        
        LightSeekState = PICKUP_GRAB;
        state_initialized = 0;
        goto start_over;
      }
      case PICKUP_GRAB:
      {
        int drive;
        int manip;

        if (!state_initialized) {
            state_initialized = 1;
            Init_Drive_Straight(30L * TICKS_PER_INCH);
        }
        state_initialized++;

        // We need to wait for the arm & drive
        drive = Drive_Ticks();
        manip = Close_Manip() || state_initialized > 60;

        if (!drive || !manip)
            break;

        if (pickup_stop)
            LightSeekState = COMPLETED;
        else        
            LightSeekState = INITIAL_SEARCH;
        state_initialized = 0;
        goto start_over;
      }
      case INITIAL_SEARCH:
      {
        if (!state_initialized) {
            state_initialized = 1;
            Arm_Auto_Init(ARM_CMD_LOW);
        }

        // We need to wait for the arm
        if (!Arm_Auto_Process())
            break;

        // Arm in position, look for light
        if (Tracking_Search_Timedout) {
            // We've looked a bit for it; no good
            LightSeekState = DRIVE_CURVE_SEARCH;
            state_initialized = 0;
            goto start_over;
        }

        if (camera_n_confidence < CONFIDENCE_COUNT)
            break;

        // We got one or more light targets in view!

        setMaxDistFromTilt();

        // One light or two?
        if ((int)GT_Packet_Data.x2 - (int)GT_Packet_Data.x1 > TWO_LIGHT_WIDTH) {
            LightSeekState = DRIVE_CURVE_SEARCH;
        } else {
            LightSeekState = GUIDED_DRIVE;
        }

        state_initialized = 0;
        goto start_over;
      }


      case DRIVE_CURVE_SEARCH:
      {
        // falls..... for now
      }


      case DRIVE_STRAIGHT_SEARCH:
      {
        if (!state_initialized) {
            state_initialized = 1;
            Init_Drive_Straight_Limited(max_search_dist,
                                   SEARCH_DRIVE_PWM_MAX);
        }

        // Did we find the light yet?
        if (GT_Packet_Data.confidence >= CONFIDENCE_THRESHOLD_DEFAULT) {
            state_initialized = 0;
            LightSeekState = GUIDED_DRIVE;
            setMaxDistFromTilt();
            pwm_left_drive = pwm_right_drive = 127;
            goto start_over;
        }

        // If we go to far, just quit before we kill someone
        if (Drive_Ticks()) {
            state_initialized = 0;
            LightSeekState = COMPLETED;
        }
        
        break;
      }

      case GUIDED_DRIVE:
      {
        if (GT_Packet_Data.confidence < CONFIDENCE_THRESHOLD_DEFAULT) {
            // Lost the light.
            LightSeekState = DRIVE_STRAIGHT_SEARCH;
            goto start_over;
        }

        if (isAtApproachDist()) {
            LightSeekState = ARM_TO_SCORE;
            pwm_left_drive = pwm_right_drive = 127;
            goto start_over;
        }
        
        setMaxDistFromTilt();
        setDriveFromPan();
        break;
      }

      case ARM_TO_SCORE:
      {
        if (!state_initialized) {
            state_initialized = 1;
            Arm_Auto_Init(score_height);
        }

        // We need to wait for the arm
        if (!Arm_Auto_Process())
            break;

        state_initialized = 0;
        LightSeekState = DRIVE_TO_SCORE;
        goto start_over;
      }

      case DRIVE_TO_SCORE:
      {
        if (!state_initialized) {
            state_initialized = 1;
            Reset_Left_Encoder_Count();
            Reset_Right_Encoder_Count();
        }

        // See if we've gone the approach distance
        if (Get_Left_Encoder_Count() >= approach_dist||
              Get_Right_Encoder_Count() >= approach_dist) {
            pwm_left_drive = pwm_right_drive = 127;
            state_initialized = 0;
            LightSeekState = RELEASE_TUBE;
            goto start_over;
        }

        if (camera_n_confidence < CONFIDENCE_THRESHOLD_DEFAULT) {
            // We don't have a light source
            // Just drive straight
            setDriveStraight();
        } else {
            // We've got a stable light source.
            setDriveFromPan();
        }
        break;
      }

      case RELEASE_TUBE:
      {
        int aret;
        if (!state_initialized) {
            state_initialized = 1;
            Arm_Auto_Init(release_height);
        }

#if 0
        // We need to wait for the arm
        if (!Arm_Auto_Process()) {
            Open_Manip();
            break;
        }
#else
        state_initialized++;
        if (!Open_Manip() && state_initialized < 60)
            break;
#endif

        Stop_Manip();
        state_initialized = 0;
        LightSeekState = BACK_OFF;
        goto start_over;
      }

      case BACK_OFF:
      {
        if (!state_initialized) {
            state_initialized = 1;
            Init_Drive_Straight(-2L * TICKS_PER_FOOT);
        }

        if (Drive_Ticks()) {
            state_initialized = 0;
            LightSeekState = COMPLETED;
        }

        break;
      }

      case COMPLETED:
        break;
    }

}

static void setMaxDistFromTilt(void)
{
    // XXX - implement me
    max_search_dist = 2L * TICKS_PER_FOOT;
}

static int isAtApproachDist(void)
{
    // lh = light height = 113.5"
    // ch = cam lense height = 32"
    // ad = desired approach distance (cam to light)
    // tilt angle = atan((lh - ch) / ad)
    // tilt pwm = tilt angle * (1.86 pwm/deg)
    // Current value (86+) based on ad == 78"
    // or... not -- based on an on-field measurement
// 96
    if (pwm_camera_tilt > 94 + TILT_CENTER_PWM_DEFAULT )
        return 1;
    else
        return 0;
}


// pan = 172   tilt = 240
static void setDriveFromPan(void)
{
    int delta = (int)pwm_camera_pan - (int)165;//(int)(PAN_CENTER_PWM_DEFAULT);
    if (delta >= 0) {
        if (delta > PAN_DRIVE_ADJ_MAX)
            delta = PAN_DRIVE_ADJ_MAX;

        // To the right
        pwm_left_drive = (unsigned char)
                      (127 + BASE_DRIVE_LEFT + delta);
        pwm_right_drive = 127 + BASE_DRIVE_RIGHT;
    } else {
        // To the left
        if (delta < -PAN_DRIVE_ADJ_MAX)
            delta = -PAN_DRIVE_ADJ_MAX;
        
        pwm_left_drive = 127 + BASE_DRIVE_LEFT;
        pwm_right_drive = (unsigned char)
                      (127 + BASE_DRIVE_RIGHT - delta);
    }
}

static void setDriveStraight(void)
{
    pwm_left_drive = 127 + BASE_DRIVE_LEFT;
    pwm_right_drive = 127 + BASE_DRIVE_RIGHT;    
}
