/*******************************************************************************
*
*	TITLE:		motion.c 
*
*	VERSION:	0.1                           
*
*	DATE:		01-25-2007
*
*	AUTHOR:		Team 1511
*
*	DESCRIPTION:	This file contains functions for autonomously driving the wheels.
********************************************************************************
*
*	CHANGE LOG:
*
*	DATE         REV  DESCRIPTION
*	-----------  ---  ----------------------------------------------------------
*	25-Jan-2007  0.1  Basic (working/tested) Ramp_Drive functions.
*
*******************************************************************************/


#include <stdio.h>
#include "motion.h"
#include "iomap.h"
#include "encoder.h"

// Default pwm value limit for PID-based driving
#define DEFAULT_DRIVE_MAX 100

static void Init_Drive_PIDData_Struct(PIDData* control, unsigned char* pwm);

static PIDData pid_left;
static PIDData pid_right;

void Motion_Init(void)
{
  Init_Drive_PIDData_Struct(&pid_left, &pwm_left_drive);
  Init_Drive_PIDData_Struct(&pid_right, &pwm_right_drive);
}

void Init_RampDrive_Struct(RampDrive* motor, unsigned char* pwm)
{
	motor->pwm = pwm;
	motor->targetSpeed = 127;
	motor->curDelta = 0;
	motor->prevSpeed = 127;
	motor->prevTarget = 127;
}

void Ramp_Drive(RampDrive* motor)
{
	unsigned char curSpeed = *(motor->pwm);
	int newSpeed;
	
	if ((motor->prevTarget != motor->targetSpeed) || (motor->curDelta == 0))	// Getting called first time with new speed, or we're already done
	{
		if ((motor->targetSpeed != motor->prevSpeed) && (motor->targetSpeed != curSpeed))	// Begin ramping if we're not already there
		{
			motor->prevSpeed = curSpeed;
			motor->prevTarget = motor->targetSpeed;
			if (motor->targetSpeed < curSpeed)
				motor->curDelta = -1;
			else
				motor->curDelta = 1;
			*(motor->pwm) += motor->curDelta;
		}
	}
	else
	{
		if (curSpeed == motor->targetSpeed)
		{
			motor->curDelta = 0;	// Done ramping
			return;
		}
		else
		{
			if (motor->curDelta > 0)
			{
				++(motor->curDelta);
				newSpeed = (int) curSpeed + motor->curDelta;	// Using int to prevent overflow
				if (newSpeed >= (int) motor->targetSpeed)
				{
					*(motor->pwm) = motor->targetSpeed;	// We're done.
					return;
				}
			}
			else
			{
				--(motor->curDelta);
				newSpeed = (int) curSpeed + motor->curDelta;	// Using int to prevent overflow
				if (newSpeed <= (int) motor->targetSpeed)
				{
					*(motor->pwm) = motor->targetSpeed;	// We're done.
					return;
				}
			}
		}
		*(motor->pwm) += motor->curDelta;	// No overflow, we can just add
	}
}

void Init_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  control->target = 0;
  control->pwm = pwm;
  control->int_error = 0;
  control->prev_error = 0;
  control->tolerance = PID_NO_OSCILLATE;
  control->kp_num = 0;
  control->kp_den = 1;
  control->ki_num = 0;
  control->ki_den = 1;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 255;
  control->i_out_max = 255;
  control->output_min = 0;
  control->output_max = 255;
}

static void Init_Drive_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  Init_PIDData_Struct(control, pwm);
  control->kp_num = 1;
  control->kp_den = 16;
  control->ki_num = 5;
  control->ki_den = 1000;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 40;
  control->i_out_max = 30;
  control->output_min = 20;
  control->output_max = DEFAULT_DRIVE_MAX;
}



void Set_PIDData_Target(PIDData* control, long new_target)
{
  if ((control->tolerance == PID_NO_OSCILLATE) ||
      ((control->target > new_target) && (control->target - new_target > control->tolerance)) || 
      ((control->target < new_target) && (new_target - control->target > control->tolerance)))
    control->int_error = 0;
  control->target = new_target;
}


unsigned char PID_Control(PIDData* control, long current)
{
  if (! disabled_mode)
  {
    long error;
    long delta_error;
    int p_out, i_out, d_out, output;

    // Compute the error & delta error    
    error = control->target - current;
    delta_error = control->prev_error - error;

    // Save this error for next delta computation
    control->prev_error = error;

    // See if we've reached our destination; if so, stop.    
    if (control->tolerance == PID_NO_OSCILLATE) {
      if (((control->target > 0) && (error < 0)) || ((control->target < 0) && (error > 0))) {
        *(control->pwm) = 127;
        control->int_error = 0;
        return 1;
      }
    } else if (error < control->tolerance && error > -control->tolerance) {
      *(control->pwm) = 127;
      control->int_error = 0;
      return 1;
    }

    // Add up the integral component, and clamp it
    control->int_error += error;
    if (control->int_error > control->int_err_max)
      control->int_error = control->int_err_max;
    else if (control->int_error < -1 * control->int_err_max)
      control->int_error = -1 * control->int_err_max;

    // Compute p component, and clamp it    
    p_out = (int) (error * control->kp_num / control->kp_den);
    if (p_out > control->p_out_max)
      p_out = control->p_out_max;
    else if (p_out < -1 * control->p_out_max)
      p_out = -1 * control->p_out_max;

    // Compute i component, and clamp it
    i_out = (int) (control->int_error * control->ki_num / control->ki_den);
    if (i_out > control->i_out_max)
      i_out = control->i_out_max;
    else if (i_out < -1 * control->i_out_max)
      i_out = -1 * control->i_out_max;

    // Compute d component
    d_out = (int) (delta_error * control->kd_num / control->kd_den);
    
    // Sum PID components
    output = p_out + i_out + d_out;

    // Clamp to minimum value
    if (output > 0 && output < control->output_min)
      output = control->output_min;
    else if (output < 0 && output > -1 * control->output_min)
      output = -1 * control->output_min;

    if (output > control->output_max)
      output = control->output_max;
    else if (output < -1 * control->output_max)
      output = -1 * control->output_max;
    
    *(control->pwm) = (unsigned char) (output + 127);
     
    return 0;
  }
  return 0;
}

unsigned char Drive_Ticks(void)
{
  unsigned char returnVal;
  long lcount, rcount;
  int lret, rret;
  int addition;

  lcount = Get_Left_Encoder_Count();
  rcount = Get_Right_Encoder_Count();

  if (pid_left.target >= 0)
  {
    addition = 15;
    lret = PID_Control(&pid_left, lcount);
    rret = PID_Control(&pid_right, rcount);
  }
  else
  {
    addition = -15;
    lret = PID_Control(&pid_left, -lcount);
    rret = PID_Control(&pid_right, -rcount);
  }    

  if (!lret && !rret) {
    long diff = lcount - rcount;
    if (diff > 25) {
      int x = *(pid_right.pwm);
      x += addition;
      *(pid_right.pwm) = (unsigned char)x;
    } else if (diff < -25) {
      int x = *(pid_left.pwm);
      x += addition;
      *(pid_left.pwm) = (unsigned char)x;
    }
  }

  return lret && rret;
}

unsigned char Turn_Ticks(void)
{
  long left_ticks, right_ticks;
  int lret, rret;
  
  left_ticks = Get_Left_Encoder_Count();
  right_ticks = Get_Right_Encoder_Count();

  lret = PID_Control(&pid_left, pid_left.target > 0 ? -left_ticks : left_ticks);
  rret = PID_Control(&pid_right, pid_right.target > 0 ? right_ticks : -right_ticks);

  return lret || rret;
         
}
  
void Init_Drive_Turn(long ticks)
{
  Reset_Left_Encoder_Count();
  Reset_Right_Encoder_Count();
  
  Set_PIDData_Target(&pid_left, -ticks);
  Set_PIDData_Target(&pid_right, ticks);

  pid_left.output_max = DEFAULT_DRIVE_MAX;
  pid_right.output_max = DEFAULT_DRIVE_MAX;
}

void Init_Drive_Straight(long ticks)
{
  Init_Drive_Straight_Limited(ticks, DEFAULT_DRIVE_MAX);
}

void Init_Drive_Straight_Limited(long ticks, int max_power)
{
  Reset_Left_Encoder_Count();
  Reset_Right_Encoder_Count();
  
  Set_PIDData_Target(&pid_left, ticks);
  Set_PIDData_Target(&pid_right, ticks);

  pid_left.output_max = max_power;
  pid_right.output_max = max_power;
}
/*
* wrote by jim bossert 
* Function name:  void Auto_Drive (int left_ratio, int right_ratio, int target_speed)
*
* Function's purpose is that offense.c to use as a way to drive toward the green tarrget 
* but could used as straight line too
*
* passed int left_ratio, int right_ratio, int target_speed
*
* return void
*
*/

// this variables are used as class local not global or this function local
int buffer_left = 0;
int buffer_right = 0;
int gain_correct_left = 0;
int gain_correct_right = 0;


void Auto_Drive (float left_ratio, float right_ratio, int target_speed)
{
  


  if (left_ratio == right_ratio)
    {
      if ((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left == (Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)
      {
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;
      
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if ((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left <= (Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)
      {
       gain_correct_left++;
       pwm_left_drive= target_speed+gain_correct_left;
       pwm_right_drive= target_speed+gain_correct_right;   
       
       buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
       buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if ((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left >= (Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)
      {
       gain_correct_left--;
       pwm_left_drive= target_speed+gain_correct_left;
       pwm_right_drive= target_speed+gain_correct_right;
        
       buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
       buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);  
      }
    }

  else if (left_ratio < right_ratio)
    {
      if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left)*right_ratio == ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)*left_ratio)
      {
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;

      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left)*right_ratio <= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)*left_ratio)
      {
      gain_correct_left++;
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;   
       
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left)*right_ratio >= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)*left_ratio)
      {
      gain_correct_left--;
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;   
      
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
    }
  else if (left_ratio > right_ratio)
    {
      if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left)*right_ratio == ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)*left_ratio)
      {
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;

      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left)*right_ratio <= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)*left_ratio)
      {
      gain_correct_right--;
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;   
       
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left)*right_ratio >= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right)*left_ratio)
      {
      gain_correct_right++;
      pwm_left_drive= target_speed+gain_correct_left;
      pwm_right_drive= target_speed+gain_correct_right;   
      
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
    }
    if (target_speed+gain_correct_left<0)
    {
     pwm_left_drive=0;
    }
    else if (target_speed+gain_correct_left>255)
    {
     pwm_left_drive=255;
    }
    if (target_speed+gain_correct_right<0)
    {
     pwm_left_drive=0;
    }
    else if (target_speed+gain_correct_right>255)
    {
     pwm_left_drive=255;
    }


  return;
}
void Reset_Buffers ()
{
  
  buffer_left=0;
  buffer_right=0;
  
  return;
}
void Reset_gain_correct ()
{
  
  gain_correct_left=0;
  gain_correct_right=0;
  
  return;
}
/*
*   wrote by jim bossert
*   it's purpose are to turn a robot by clockwise
*   by turn left pwm forward and right pwm backward
*
*/
void clockwise_turn (int speed_target)
{
  int target_speed= speed_target-127;

  if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left) == ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right))
      {
      pwm_left_drive= 127+target_speed+gain_correct_left;
      pwm_right_drive= 127+(-1*(target_speed+gain_correct_right));

      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left) <= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right))
      {
      gain_correct_left++;
      pwm_left_drive= 127+target_speed+gain_correct_left;
      pwm_right_drive= 127+(-1*(target_speed+gain_correct_right));   
       
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left) >= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right))
      {
      gain_correct_left--;
      pwm_left_drive=127+target_speed+gain_correct_left;
      pwm_right_drive= 127+(-1*(target_speed+gain_correct_right));   
      
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }

    if (127+target_speed+gain_correct_left<0)
    {
     pwm_left_drive=0;
    }
    else if (127+target_speed+gain_correct_left>255)
    {
     pwm_left_drive=255;
    }
    if (127+(-1*(target_speed+gain_correct_right))<0)
    {
     pwm_left_drive=0;
    }
    else if (127+(-1*(target_speed+gain_correct_right))>255)
    {
     pwm_left_drive=255;
    }

  return;
}
/*
*   wrote by jim bossert
*   it's purpose are to turn a robot by counterclockwise
*   by turn left pwm backward and right pwm forward
*
*/
void counter_clockwise_turn (int speed_target)
{

  int target_speed= speed_target-127;

if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left) == ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right))
      {
      pwm_left_drive= 127+(-1*(target_speed+gain_correct_left));
      pwm_right_drive= 127+(target_speed+gain_correct_right);

      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left) <= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right))
      {
      gain_correct_right--;
      pwm_left_drive= 127+(-1*(target_speed+gain_correct_left));
      pwm_right_drive= 127+(target_speed+gain_correct_right);   
       
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }
      else if (((Get_Left_Encoder_Count()/TICKS_PER_INCH)-buffer_left) >= ((Get_Right_Encoder_Count()/TICKS_PER_INCH)-buffer_right))
      {
      gain_correct_right++;
      pwm_left_drive= 127+(-1*(target_speed+gain_correct_left));
      pwm_right_drive= 127+(target_speed+gain_correct_right);   
      
      buffer_left = (Get_Left_Encoder_Count()/TICKS_PER_INCH);
      buffer_right = (Get_Right_Encoder_Count()/TICKS_PER_INCH);
      }  

    if (127+target_speed+gain_correct_left<0)
    {
     pwm_left_drive=0;
    }
    else if (127+target_speed+gain_correct_left>255)
    {
     pwm_left_drive=255;
    }
    if (127+(-1*(target_speed+gain_correct_right))<0)
    {
     pwm_left_drive=0;
    }
    else if (127+(-1*(target_speed+gain_correct_right))>255)
    {
     pwm_left_drive=255;
    }

  return;
}
